import java.util.*;

/**
 *Klauseln sind die oberste Einheit der Sprache.
 *
 *z.B. sind Regeln und Fakten Klauseln:<p>
 * <code>p(a, 8, g(Z)) ---> b(Z,8).</code> <p>ist eine Klausel. Ebenso<p>
 * <code>---> doof(harvey).</code> <p>und <p><code>schlau(harvey)---></code> .
 *<p>
 *Klauseln drfen rechts des Pfeiles keine Variablen haben, die nicht
 *auch links des
 *Pfeiles bereits vorkommen; denn sonst wre die Klausel - selbst wenn die
 *linke Seite voll instanziiert wre - nicht grund (ground).
 *<p>
 *Man kann den Klauseln einen Namen geben, indem man einen Namen vor die 
 *Klausel schreibt und durch '::' von ihr trennt.
 *<p>
 *Siehe auch <a href="SatchParserBNF.html"><b>SatchParserBNF.html</b></a>.
 */
class Clause{

  private Body body;
  private Head head;
  private String identifier;

  /**
   *Der identifier-String dient zur besseren Orientierung. Man kann ihn
   *in der Specificationsdatei als <identifier>:: vor die Clausel schreiben.
   *Schreibt man keinen eigenen identifier, wird einer generiert.
   */
  Clause(String identifier, Body b, Head h){
    body = b; head = h;
    this.identifier = identifier;
  }

  Clause(String identifier, Body b){
    body = b; head = null;
    this.identifier = identifier;
  }

  Clause(String identifier, Head h){
    body = null ; head = h;
    this.identifier = identifier;
  }
  

  /**
   *Gibt den in der Spezifikation angegebenen, bzw. generierten Namen der
   *Klausel zurck.
   */
  String getIdentifier(){
    return identifier;
  }


  void setIdentifier(String identifier){
    this.identifier = identifier;
  }

  /**
   *Gibt <code>null</code> zurck, falls diese Klausel keinen 
   *<code>Body</code> hat.
   */
  Body getBody(){
    return body; 
  }
 
  /**
   *Gibt <code>null</code> zurck, falls diese Klausel keinen 
   *<code>Head</code> hat.
   */
  Head getHead(){
    return head; 
  }

  Clause copy(){

    Body bodyKlon = null;
    Head headKlon = null;

    if(body != null){
      body.cloneYourValue();
      bodyKlon = body.clonedValue();
    }
    if(head != null){
      head.cloneYourValue();
      headKlon = head.clonedValue();
    }
    
    return new Clause(identifier, bodyKlon, headKlon);
  }


  /**
   *Diese Methode wird vom Parser nach dem Parsen aufgerufen, und stellt 
   *sicher, da an jeder Stelle der Clausel, an der zwei gleichlautende 
   *Variablen stehen, das selbe Var-Objekt steht.<p>
   *
   *Dabei wird ein Vector angelegt (falls keiner bergeben wurde), in den 
   *alle Variablen-Objekte hineingeschrieben werden, die bisher vorkamen.
   *Dieser Vector wird an alle in dieser Clausel vorkommenden Clauselteile 
   *bergeben, damit alle Variablen, die in den untergeordneten Objekten
   *vorkommen, durch die in anderen Teilen dieser Clausel vorkommenden 
   *Variablen-Objekte ersetzt werden.<p>
   *
   *Ist die Clausel nicht ground, wird eine ClauseNotGroundException geworfen.
   */
  public void internVars() throws ClauseNotGroundException{
    Vector v = new Vector();
    
    if(body != null)
      v = body.internVars(v);

    if(head != null){
      int varsCount = v.size();//v enthlt alle Variblen der Klausel
      v = head.internVars(v);  //sollten jetzt noch neue, unbekannte dazukommen,

      if(varsCount < v.size()) //ist die Klausel nicht ground
	throw new ClauseNotGroundException("Clause "+identifier+" not ground! "+
       "\n\t(= there are some Vars on the right side, which are not on the left!)");
    }
  }
  
  /**
   *Wenn die Clausel einen Body, aber keinen Head hat, ist sie im Tableaux 
   *schon beim Start als 'wahr' anzusehen. Wre equivalent zu '<code>hasABody
   *</code>'.
   */
  boolean isTrue(){ 
    return body == null;//Da Clauseln ohne Body und Head vermieden werden;-)
  }

  /**
   *Eine Clausel ist ein Constraint, wenn sie einen Body, aber keinen Head hat.
   */
  boolean isConstraint(){
    return head == null;
  }

  public boolean equals(Object o){
    if(o instanceof Clause){
      Clause c = (Clause) o;
      if(c.getHead() == null){
	if(head == null && c.getBody().equals(body))
	  return true;
	return false;
      }
      else if(c.getBody() == null){
	if(body == null && c.getHead().equals(head))
	  return true;
	return false;
      }else if(c.getHead().equals(head) & c.getBody().equals(body))
	return true;
      else
	return false;
    }
    return false;
  }


  public String toString(){
      return (body == null ? "true" : identifier+"__"+body.toString())+
                           " ---> "
                           +(head == null ? "false" : head.toString());
  }
}

/**
 *Wird ausgelst, wenn die Klausel nicht 'ground' ist.
 *
 *  @see Clause
 */
class ClauseNotGroundException extends Exception{
  public ClauseNotGroundException(){super();}
  public ClauseNotGroundException(String s){super(s);}
}
